const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

const kNotificationSeqURI = "http://celtx.com/NS/v1/NotificationSeq";


function nsCeltxService () {
  var rdfsvc = Cc["@mozilla.org/rdf/rdf-service;1"]
    .getService(Ci.nsIRDFService);
  var rdfcu = Cc["@mozilla.org/rdf/container-utils;1"]
    .getService(Ci.nsIRDFContainerUtils);

  var ds = Cc["@mozilla.org/file/directory_service;1"]
    .getService(Ci.nsIProperties);
  var dir = ds.get("ProfD", Ci.nsIFile);
  dir.append("temp");
  if (! dir.exists() || ! (dir.isReadable() && dir.isWritable()))
    dir.createUnique(1, 0700);
  this._tempDir = dir;

  // TODO: Proper initialization
  this._notificationURL = this.AUTH_SCHEME + "://" + this.PUBLISH_SERVER
      + "/app/notice/list";
  this._notificationInterval = 15000; // every 15 seconds

  this._notificationds = null;
  this._notifications = null;
  var notifycache = ds.get("ProfD", Ci.nsIFile);
  notifycache.append("celtx_notifications.rdf");
  if (notifycache.exists()) {
    try {
      var ios = Cc["@mozilla.org/network/io-service;1"]
        .getService(Ci.nsIIOService);
      this._notificationds = rdfsvc.GetDataSourceBlocking(
        ios.newFileURI(notifycache).spec);
    }
    catch (ex) {
      this._notificationds = null;
      dump("*** nsCeltxService: Unable to load notifycache\n" + ex + "\n");
    }
  }
  if (! this._notificationds) {
    this._notificationds =
      Cc["@mozilla.org/rdf/datasource;1?name=in-memory-datasource"]
      .createInstance(Ci.nsIRDFDataSource);
  }

  var seqres = rdfsvc.GetResource(kNotificationSeqURI);
  var seq = null;
  if (! rdfcu.IsSeq(this._notificationds, seqres)) {
    seq = rdfcu.MakeSeq(this._notificationds, seqres);
  }
  else {
    // Fixup for empty rdf:Seq
    var nextval = rdfsvc.GetResource(
      "http://www.w3.org/1999/02/22-rdf-syntax-ns#nextVal");
    var one = rdfsvc.GetLiteral("1");
    if (! this._notificationds.hasArcOut(seqres, nextval))
      this._notificationds.Assert(seqres, nextval, one, true);
    seq = Cc["@mozilla.org/rdf/container;1"].createInstance(Ci.nsIRDFContainer);
    seq.Init(this._notificationds, seqres);
  }
  this._notifications = seq;

  var os = Cc["@mozilla.org/observer-service;1"]
    .getService(Ci.nsIObserverService);
  os.addObserver(this, "quit-application", false);
  os.addObserver(this, "celtx:login-status-changed", false);

  this.startup();
}

nsCeltxService.prototype = {
  QueryInterface: function cxsvc_QI(iid) {
    if (iid.equals(Ci.nsISupports) ||
        iid.equals(Ci.nsIObserver) ||
        iid.equals(Ci.nsICeltxService))
      return this
    throw Cr.NS_ERROR_NO_INTERFACE;
  },


  _offline: true,
  _loggedIn: false,
  _autologinChecked: false,
  _suspendNotifications: false,
  _notificationInterval: null,
  _notificationURL: null,
  _notificationTimer: null,

  get VERSION ()      { return "1.0"; },
  get AUTH_REALM ()   { return "celtx"; },
  get AUTH_TYPE ()    { return "basic"; },
  get AUTH_PORT ()    { return 80;      },
  get AUTH_SCHEME ()  { return "https"; },
  get AUTH_DOMAIN ()  { return null;    },
  get AUTH_PATH ()    { return "/";     },

  get PUBLISH_SERVER () {
    var ps = Cc["@mozilla.org/preferences-service;1"]
      .getService(Ci.nsIPrefService);
    return ps.getBranch("celtx.server.").getCharPref("selection");
  },
  get AUTH_URL () {
    return this.AUTH_SCHEME + "://" + this.PUBLISH_SERVER + "/app/test-auth";
  },
  get STARTUP_URL () {
    return this.AUTH_SCHEME + "://" + this.PUBLISH_SERVER + "/pub/startup";
  },


  startup: function startup () {
    var request = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"]
      .createInstance(Components.interfaces.nsIJSXMLHttpRequest);
    request.onload = function () {
      var cxsvc = Components.classes["@celtx.com/celtx-service;1"]
        .getService(Components.interfaces.nsICeltxService);
      if (this.status >= 200 && this.status < 300)
        cxsvc.offline = false;
      else
        cxsvc.offline = true;
    };
    request.onerror = function () {
      var cxsvc = Components.classes["@celtx.com/celtx-service;1"]
        .getService(Components.interfaces.nsICeltxService);
      cxsvc.offline = true;
    };
    request = request.QueryInterface(Ci.nsIXMLHttpRequest);
    request.open("GET", this.STARTUP_URL + "/" + this.VERSION, true);
    var channel = request.channel;
    var prompt = new SuppressedAuthPrompt(channel.notificationCallbacks);
    channel.notificationCallbacks = prompt;
    request.send(null);
  },


  get notificationCount () {
    return this._notifications.GetCount();
  },


  peekNotification: function peekNotification () {
    if (this.notificationCount <= 0)
      return null;
    try {
      var cu = Cc["@mozilla.org/rdf/container-utils;1"]
        .getService(Ci.nsIRDFContainerUtils);
      var firstarc = cu.IndexToOrdinalResource(1);
      var res = this._notificationds.GetTarget(this._notifications.Resource,
        firstarc, true);
      if (! res)
        return null;
      res = res.QueryInterface(Ci.nsIRDFResource);
      return new Notification(this._notificationds, res);
    }
    catch (ex) {
      dump("*** nsCeltxService.peekNotification: " + ex + "\n");
      return null;
    }
  },


  nextNotification: function nextNotification () {
    if (this.notificationCount <= 0)
      return null;
    try {
      var cu = Cc["@mozilla.org/rdf/container-utils;1"]
        .getService(Ci.nsIRDFContainerUtils);
      var firstarc = cu.IndexToOrdinalResource(1);
      var res = this._notificationds.GetTarget(this._notifications.Resource,
        firstarc, true);
      if (! res)
        return null;
      res = res.QueryInterface(Ci.nsIRDFResource);
      var notification = new Notification(this._notificationds, res);
      this._notifications.RemoveElementAt(1, true);
      var obssvc = Cc["@mozilla.org/observer-service;1"]
        .getService(Ci.nsIObserverService);
      obssvc.notifyObservers(this, "celtx:notification-count-changed",
        this.notificationCount);
      return notification;
    }
    catch (ex) {
      dump("*** nsCeltxService.nextNotification: " + ex + "\n");
      return null;
    }
  },


  shutdown: function shutdown () {
    try {
      this._tempDir.remove(true);
    }
    catch (ex) {
      dump("*** nsCeltxService.shutdown: " + ex + "\n");
    }

    if (this._notificationTimer) {
      this._notificationTimer.cancel();
      this._notificationTimer = null;
    }

    try {
      var ds = Cc["@mozilla.org/file/directory_service;1"]
        .getService(Ci.nsIProperties);
      var notifycache = ds.get("ProfD", Ci.nsIFile);
      notifycache.append("celtx_notifications.rdf");

      var atomsvc = Cc["@mozilla.org/atom-service;1"]
        .getService(Ci.nsIAtomService);
      var serializer = Cc["@mozilla.org/rdf/xml-serializer;1"]
        .createInstance(Ci.nsIRDFXMLSerializer);
      serializer.addNameSpace(atomsvc.getAtom("cx"), "http://celtx.com/NS/v1/");
      serializer.addNameSpace(atomsvc.getAtom("dc"),
        "http://purl.org/dc/elements/1.1/");
      serializer.init(this._notificationds);
      var fs = Cc["@mozilla.org/network/file-output-stream;1"]
        .createInstance(Ci.nsIFileOutputStream);
      // PR_RDWR | PR_CREATE_FILE | PR_TRUNCATE
      fs.init(notifycache, 0x04 | 0x08 | 0x20, 0600, 0);
      var os = Cc["@mozilla.org/network/buffered-output-stream;1"]
        .createInstance(Ci.nsIBufferedOutputStream);
      os.init(fs, 4096);
      serializer = serializer.QueryInterface(Ci.nsIRDFXMLSource);
      serializer.Serialize(os);
      os.close();
      this._notifications = null;
      this._notificationds = null;
    }
    catch (ex) {
      dump("*** nsCeltxService.shutdown: " + ex + "\n");
    }

    var os = Cc["@mozilla.org/observer-service;1"]
      .getService(Ci.nsIObserverService);
    os.removeObserver(this, "quit-application");
    os.removeObserver(this, "celtx:login-status-changed");
  },


  observe: function (subject, topic, data) {
    if (topic == "quit-application") {
      this.shutdown();
    }
    else if (topic == "celtx:login-status-changed") {
      // If we're logging in or out, the timer is invalid.
      if (this._notificationTimer) {
        this._notificationTimer.cancel();
        this._notificationTimer = null;
      }
      if (data == "loggedin" && ! this._suspendNotifications)
        this.notify(null);
    }
  },


  get tempDirSpec () {
    var ios = Cc["@mozilla.org/network/io-service;1"]
      .getService(Ci.nsIIOService);
    return ios.newFileURI(this._tempDir).spec;
  },


  get offline () {
    return this._offline;
  },


  set offline (val) {
    if (this._offline != val) {
      this._offline = val;
      var os = Cc["@mozilla.org/observer-service;1"]
        .getService(Ci.nsIObserverService);
      os.notifyObservers(this, "celtx:network-status-changed",
        val ? "offline" : "online");
    }
    return val;
  },


  get loggedIn () {
    return this._loggedIn;
  },


  loginSuccessful: function cxsvc_loginSuccessful () {
    this._loggedIn = true;
  },


  get username () {
    if (! this.loggedIn)
      return null;

    var authMgr = Cc["@mozilla.org/network/http-auth-manager;1"]
      .getService(Ci.nsIHttpAuthManager);
    var cxsvc = Cc["@celtx.com/celtx-service;1"].getService(Ci.nsICeltxService);
    var rv = { domain: {}, username: {}, password: {} };
    try {
      authMgr.getAuthIdentity(cxsvc.AUTH_SCHEME, cxsvc.PUBLISH_SERVER,
        cxsvc.AUTH_PORT, cxsvc.AUTH_TYPE, cxsvc.AUTH_REALM,
        cxsvc.AUTH_PATH, rv.domain, rv.username, rv.password);
      return rv.username.value;
    }
    catch (ex) {
      return null;
    }
  },


  checkAutoLogin : function cxsvc_checkAutoLogin () {
    if (this._autologinChecked)
      return;

    this._autologinChecked = true;

    var ps = Cc["@mozilla.org/preferences-service;1"]
      .getService(Ci.nsIPrefService).getBranch("celtx.");
    if (! ps.getBoolPref("user.loginOnStartup"))
      return;

    try {
      var userid = ps.getCharPref("user.id");
      var pass = ps.getCharPref("user.encpassword");
      pass = base64_decodew(pass);
      this.loginAs(userid, pass, null, {onLogin: function (success) {}}, false);
    }
    catch (ex) {
      dump("*** nsCeltxService: " + ex + "\n");
    }
  },


  // |service| is unused, but consumers expect it
  login: function cxsvc_login (service, observer, reattempt, win) {
    if (this.loggedIn) {
      observer.onLogin(true);
      return;
    }

    // var request = new XMLHttpRequest();
    var request = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"]
      .createInstance(Ci.nsIJSXMLHttpRequest);
    request.onload  = function () {
      // var authMgr = getAuthManager();
      var authMgr = Cc["@mozilla.org/network/http-auth-manager;1"]
        .getService(Ci.nsIHttpAuthManager);
      if (request.status == 401) {
        // Not Authorized
        dump("*** nsCeltxService.login: 401 not authorized\n");
        authMgr.clearAll();
        // Retry
        this.login(service, observer, true, win);
      }
      else if (request.status != 200) {
        dump("*** nsCeltxService.login: status: " + request.status + "\n");
        authMgr.clearAll();
        if (observer)
          observer.onLogin(false);
      }
      else {
        var cxsvc = Cc["@celtx.com/celtx-service;1"]
          .getService(Ci.nsICeltxService);
        cxsvc.loginSuccessful();
        var obssvc = Cc["@mozilla.org/observer-service;1"]
          .getService(Ci.nsIObserverService);
        obssvc.notifyObservers(this, "celtx:login-status-changed", "loggedin");
        if (observer)
          observer.onLogin(true);
      }
    };
    request.onerror = function () {
      if (observer)
        observer.onLogin(false);
    };
    request.onreadystatechange = function () {
      var uninitialized = 0;
      var completed = 4;
      if (! observer)
        return;

      // If the request is aborted, it broadcasts a change to COMPLETED and
      // then silently switches to UNINITIALIZED, so we need to check on
      // a timeout if the state changes after we're notified.
      if (request.readyState == completed) {
        win.setTimeout(function () {
          if (request.readyState == uninitialized)
            observer.onLogin(false);
        }, 0);
      }
    };
    request = request.QueryInterface(Ci.nsIXMLHttpRequest);
    request.open("GET", this.AUTH_URL, true);
    var channel = request.channel;
    var prompt = new CeltxAuthPrompt(service, channel.notificationCallbacks,
                                     request, win);
    if (reattempt)
      prompt.reattempt = true;
    channel.notificationCallbacks = prompt;
    request.send(null);
  },


  loginAs: function cxsvc_loginAs (username, password, service, observer,
                                   synchronous) {
    // Trim leading and trailing white space
    username = username.replace(/^\s+/, "");
    username = username.replace(/\s+$/, "");
    if (this.loggedIn && this.username == username) {
      observer.onLogin(true);
      return;
    }

    var authMgr = Cc["@mozilla.org/network/http-auth-manager;1"]
      .getService(Ci.nsIHttpAuthManager);
    authMgr.setAuthIdentity(this.AUTH_SCHEME, this.PUBLISH_SERVER,
      this.AUTH_PORT, this.AUTH_TYPE, this.AUTH_REALM, this.AUTH_PATH,
      this.AUTH_DOMAIN, username, password);

    // var request = new XMLHttpRequest();
    var request = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"]
      .createInstance(Ci.nsIJSXMLHttpRequest);
    request.onload  = function () {
      // var authMgr = getAuthManager();
      var authMgr = Cc["@mozilla.org/network/http-auth-manager;1"]
        .getService(Ci.nsIHttpAuthManager);
      if (request.status != 200) {
        dump("*** nsCeltxService.loginAs: status: " + request.status + "\n");
        authMgr.clearAll();
        if (observer)
          observer.onLogin(false);
      }
      else {
        var cxsvc = Cc["@celtx.com/celtx-service;1"]
          .getService(Ci.nsICeltxService);
        cxsvc.loginSuccessful();
        var obssvc = Cc["@mozilla.org/observer-service;1"]
          .getService(Ci.nsIObserverService);
        obssvc.notifyObservers(this, "celtx:login-status-changed", "loggedin");
        if (observer)
          observer.onLogin(true);
      }
    };
    request.onerror = function () {
      if (observer)
        observer.onLogin(false);
    };
    request = request.QueryInterface(Ci.nsIXMLHttpRequest);
    request.open("GET", this.AUTH_URL, true);
    var channel = request.channel;
    var prompt = new SuppressedAuthPrompt(channel.notificationCallbacks);
    prompt.username = username;
    prompt.password = password;
    channel.notificationCallbacks = prompt;
    request.send(null);
  },


  logout: function cxsvc_logout () {
    this._loggedIn = false;
    var authMgr = Cc["@mozilla.org/network/http-auth-manager;1"]
      .getService(Ci.nsIHttpAuthManager);
    authMgr.clearAll();
    var obssvc = Cc["@mozilla.org/observer-service;1"]
      .getService(Ci.nsIObserverService);
    obssvc.notifyObservers(this, "celtx:login-status-changed", "loggedout");
  },


  // Check server for notifications
  notify: function notify (timer) {
    var req = Cc["@mozilla.org/xmlextras/xmlhttprequest;1"]
      .createInstance(Ci.nsIJSXMLHttpRequest);
    req.open("POST", this._notificationURL, true);
    req.setRequestHeader("Content-Type", "application/xml");
    var self = this;
    req.onload = function (event) {
      self.notify_onload (req);
    };
    req.onerror = function (event) {};
    var body = "<?xml version='1.0' encoding='UTF-8'?>\n\n"
      + "<request type='get' xml:lang='en'/>\n\n";
    req.send(body);
  },

  notify_onload: function notify_onload (req) {
    dump("--- notify_onload\n");
    if (req.status < 200 || req.status >= 300) {
      dump("*** nsCeltxService.notify: status == " + req.status + "\n");
      this._suspendNotifications = true;
      return;
    }

    var ios = Cc["@mozilla.org/network/io-service;1"]
      .getService(Ci.nsIIOService);
    var tempds = Cc["@mozilla.org/rdf/datasource;1?name=in-memory-datasource"]
      .createInstance(Ci.nsIRDFDataSource);
    var parser = Cc["@mozilla.org/rdf/xml-parser;1"]
      .createInstance(Ci.nsIRDFXMLParser);
    var rdfsvc = Cc["@mozilla.org/rdf/rdf-service;1"]
      .getService(Ci.nsIRDFService);
    var typearc = rdfsvc.GetResource(
      "http://www.w3.org/1999/02/22-rdf-syntax-ns#type");
    var cxtype = rdfsvc.GetResource(
      "http://celtx.com/NS/v1/NotificationResponse");
    var notificationsarc = rdfsvc.GetResource(
      "http://celtx.com/NS/v1/notifications");
    var intervalarc = rdfsvc.GetResource("http://celtx.com/NS/v1/pollInterval");
    var redirectarc = rdfsvc.GetResource("http://celtx.com/NS/v1/redirect");
    var notificationCount = 0;
    try {
      parser.parseString(tempds,
        ios.newURI("http://celtx.com/NS/v1/", "UTF-8", null), req.responseText);
      var responses = tempds.GetSources(typearc, cxtype, true);
      while (responses.hasMoreElements()) {
        var response = responses.getNext()
          .QueryInterface(Ci.nsIRDFResource);

        // Check for a cx:pollInterval value
        var interval = tempds.GetTarget(response, intervalarc, true);
        if (interval && (interval instanceof Ci.nsIRDFLiteral)) {
          this._notificationInterval = Number(
            interval.QueryInterface(Ci.nsIRDFLiteral).Value) * 1000;
          if (this._notificationInterval <= 0)
            this._suspendNotifications = true;
        }

        // Check for a cx:redirect value
        var redirect = tempds.GetTarget(response, redirectarc, true);
        if (redirect && (redirect instanceof Ci.nsIRDFLiteral)) {
          this._notificationURL =
            redirect.QueryInterface(Ci.nsIRDFLiteral).Value;
        }

        // Check for a cx:notifications sequence
        var seqres = tempds.GetTarget(response, notificationsarc, true);
        if (! seqres)
          continue;
        seqres = seqres.QueryInterface(Ci.nsIRDFResource);
        var seq = Cc["@mozilla.org/rdf/container;1"]
          .createInstance(Ci.nsIRDFContainer);
        seq.Init(tempds, seqres);
        var notifications = seq.GetElements();
        while (notifications.hasMoreElements()) {
          var notification = notifications.getNext()
            .QueryInterface(Ci.nsIRDFResource);
          this._notifications.AppendElement(notification);
          var arcs = tempds.ArcLabelsOut(notification);
          while (arcs.hasMoreElements()) {
            var arc = arcs.getNext().QueryInterface(Ci.nsIRDFResource);
            var target = tempds.GetTarget(notification, arc, true);
            this._notificationds.Assert(notification, arc, target, true);
          }
          ++notificationCount;
        }
      }
    }
    catch (ex) {
      dump("*** nsCeltxService.notify: " + ex + "\n");
    }

    if (notificationCount > 0) {
      var obssvc = Cc["@mozilla.org/observer-service;1"]
        .getService(Ci.nsIObserverService);
      obssvc.notifyObservers(this, "celtx:notification-count-changed",
        this.notificationCount);
    }

    if (this._suspendNotifications) {
      if (this._notificationTimer) {
        this._notificationTimer.cancel();
        this._notificationTimer = null;
      }
      return;
    }

    if (! this._notificationTimer) {
      this._notificationTimer = Cc["@mozilla.org/timer;1"]
        .createInstance(Ci.nsITimer);
      this._notificationTimer.initWithCallback(this,
        this._notificationInterval, Ci.nsITimer.TYPE_REPEATING_SLACK);
    }
    else if (this._notificationTimer.delay != this._notificationInterval) {
      this._notificationTimer.delay = this._notificationInterval;
    }
      
  }
};


var initModule = {
  ServiceCID: Components.ID("{879e5daa-510e-4e1c-9420-66ac2b7bf7a3}"),
  ServiceContractID: "@celtx.com/celtx-service;1",
  ServiceName: "Celtx Service",


  registerSelf: function (compMgr, fileSpec, location, type) {
    compMgr = compMgr.QueryInterface(Ci.nsIComponentRegistrar);
    compMgr.registerFactoryLocation(this.ServiceCID, this.ServiceName,
      this.ServiceContractID, fileSpec, location, type);
  },


  unregisterSelf: function (compMgr, fileSpec, location) {
    compMgr = compMgr.QueryInterface(Ci.nsIComponentRegistrar);
    compMgr.unregisterFactoryLocation(this.ServiceCID, fileSpec);
  },


  getClassObject: function (compMgr, cid, iid) {
    if (! cid.equals(this.ServiceCID))
      throw Cr.NS_ERROR_NO_INTERFACE;
    if (! iid.equals(Ci.nsIFactory))
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;
    return this.instanceFactory;
  },


  canUnload: function (compMgr) {
    return true;
  },


  instanceFactory: {
    createInstance: function (outer, iid) {
      if (outer != null)
        throw Cr.NS_ERROR_NO_AGGREGATION;
      return new nsCeltxService().QueryInterface(iid);
    }
  }
};


function NSGetModule (compMgr, fileSpec) {
  return initModule;
}


function CeltxAuthPrompt (service, callbacks, request, win) {
  this.service = service;
  this.notificationCallbacks = callbacks;
  this.request = request;
  this.reattempt = false;
  this.win = win;
}


CeltxAuthPrompt.prototype = {
  QueryInterface: function (iid) {
    if (iid.equals(Ci.nsIInterfaceRequestor))
      return this;
    throw Cr.NS_NOINTERFACE;
  },
  getInterface: function (iid) {
    if (iid.equals(Ci.nsIAuthPrompt) ||
        iid.equals(Ci.nsIAuthPromptProvider) ||
        iid.equals(Ci.nsIProgressEventSink))
      return this;
    throw Cr.NS_NOINTERFACE;
  },

  PROMPT_NORMAL: 0,
  PROMPT_PROXY: 1,

  getAuthPrompt: function (reason) { return this; },
  onProgress: function (req, ctx, prog, progmax) {},
  onStatus: function (req, ctx, status, statusarg) {},
  alert: function (title, text) {},
  alertCheck: function (title, text, checkmsg, checkval) {},
  confirm: function (title, text) {},
  confirmCheck: function (title, text, checkmsg, checkval) {},
  confirmEx: function (title, text, flags, b0, b1, b2,
                       checkmsg, checkval) {},
  prompt: function (title, text, realm, savepass, text, result) {},
  promptPassword: function (title, text, realm, savepass, pass) {},
  select: function (title, text, count, list, selection) {},
  promptUsernameAndPassword: function (title, text, realm, savepass,
                                       user, pass) {
    // user, pass, and checkval are inout
    var cxsvc = Cc["@celtx.com/celtx-service;1"].getService(Ci.nsICeltxService);
    var auth = {
      username: user.value,
      password: pass.value,
      service: this.service,
      server: cxsvc.PUBLISH_SERVER
    };

    if (this.reattempt)
      auth.reattempt = true;

    // var cxsvc = Cc["@celtx.com/celtx-service;1"].getService(Ci.nsICeltxService);
    // auth.server = cxsvc.PUBLISH_SERVER;

    if (this.win)
      this.win.openDialog("chrome://celtx/content/authenticate.xul", "_blank",
                          "chrome,modal,centerscreen,titlebar", auth);
    else
      window.openDialog("chrome://celtx/content/authenticate.xul",
                        "_blank",
                        // Cx.MODAL_DIALOG_FLAGS,
                        "chrome,modal,centerscreen,titlebar",
                        auth);

    if (auth.canceled) {
      this.request.abort();
      return false;
    }

    // Trim leading and trailing white space
    auth.username = auth.username.replace(/^\s+/, "");
    auth.username = auth.username.replace(/\s+$/, "");

    // var authMgr = getAuthManager();
    var authMgr = Cc["@mozilla.org/network/http-auth-manager;1"]
      .getService(Ci.nsIHttpAuthManager);
    authMgr.setAuthIdentity(cxsvc.AUTH_SCHEME, cxsvc.PUBLISH_SERVER,
      cxsvc.AUTH_PORT, cxsvc.AUTH_TYPE, cxsvc.AUTH_REALM, cxsvc.AUTH_PATH,
      cxsvc.AUTH_DOMAIN, auth.username, auth.password);

    user.value = auth.username;
    pass.value = auth.password;
    // gApp.userID = auth.username;
    return true;
  }
};


function SuppressedAuthPrompt (callbacks) {
  this.notificationCallbacks = callbacks;
}


SuppressedAuthPrompt.prototype = {
  QueryInterface: function (iid) {
    if (iid.equals(Ci.nsIInterfaceRequestor))
      return this;
    throw Cr.NS_NOINTERFACE;
  },
  getInterface: function (iid) {
    if (iid.equals(Ci.nsIPrompt) ||
        iid.equals(Ci.nsIAuthPrompt) ||
        iid.equals(Ci.nsIAuthPromptProvider) ||
        iid.equals(Ci.nsIProgressEventSink))
      return this;
    throw Cr.NS_NOINTERFACE;
  },

  PROMPT_NORMAL: 0,
  PROMPT_PROXY: 1,

  getAuthPrompt: function (reason) { return this; },
  onProgress: function (req, ctx, prog, progmax) {},
  onStatus: function (req, ctx, status, statusarg) {},

  alert: function (title, text) {},
  alertCheck: function (title, text, checkmsg, checkval) {},
  confirm: function (title, text) {},
  confirmCheck: function (title, text, checkmsg, checkval) {},
  confirmEx: function (title, text, flags, b0, b1, b2,
                       checkmsg, checkval) {},
  prompt: function (title, text, realm, savepass, text, result) {},
  promptPassword: function (title, text, realm, savepass, pass) {},
  select: function (title, text, count, list, selection) {},
  promptUsernameAndPassword: function (title, text, realm, savepass,
                                       user, pass) {
    if (this.alreadyTried)
      return false;

    user.value = this.username;
    pass.value = this.password;
    this.alreadyTried = true;
    return true;
  }
};


function base64_decodew (str) {
  var table =
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
  var result = "";
  var i = 0;
  var accum = -1;
  while (i < str.length) {
    var src = [
      table.indexOf(str.charAt(i++)), table.indexOf(str.charAt(i++)),
      table.indexOf(str.charAt(i++)), table.indexOf(str.charAt(i++))
    ];
    if (isNaN(src[0]) || isNaN(src[1]) || isNaN(src[2]) || isNaN(src[3]) ||
        src[0] < 0 || src[1] < 0 || src[2] < 0 || src[3] < 0)
      throw "String does not appear to be base64";
    // Masking guarantees any '=' will be trimmed from 0x40 to 0x00
    var dst = [
      (src[0] << 2) | ((src[1] & 0x3F) >> 4),
      ((src[1] & 0x0F) << 4) | ((src[2] & 0x3F) >> 2),
      ((src[2] & 0x03) << 6) | (src[3] & 0x3F)
    ];
    if (i % 8 == 4) {
      // First char and a half
      result += String.fromCharCode((dst[0] << 8) | dst[1]);
      if (src[3] == 64) {
        break;
      }
      accum = dst[2] << 8;
    }
    else {
      result += String.fromCharCode(accum | dst[0]);
      accum = -1;
      if (dst[1] == 0 && dst[2] == 0) {
        break;
      }
      result += String.fromCharCode((dst[1] << 8) | dst[2]);
    }
  }
  return result;
}


function Notification (ds, res) {
  var rdfsvc = Components.classes["@mozilla.org/rdf/rdf-service;1"]
    .getService(Components.interfaces.nsIRDFService);
  var NS_DC = "http://purl.org/dc/elements/1.1/";
  var ILit = Components.interfaces.nsIRDFLiteral;
  var fields = [ "type", "title", "date", "description", "source" ];

  for (var i = 0; i < fields.length; ++i) {
    var arc = rdfsvc.GetResource(NS_DC + fields[i]);
    var lit = ds.GetTarget(res, arc, true);
    if (lit instanceof ILit)
      this["_" + fields[i]] = lit.QueryInterface(ILit).Value;
    else
      this["_" + fields[i]] = null;
  }
}


Notification.prototype = {
  QueryInterface: function (iid) {
    if (iid.equals(Components.interfaces.nsISupports) ||
        iid.equals(Components.interfaces.nsICeltxNotification))
      return this;
    throw Components.results.NS_NOINTERFACE;
  },

  get type () { return this._type; },
  set type (val) { return this._type = val; },

  get title () { return this._title; },
  set title (val) { return this._title = val; },

  get date () { return this._date; },
  set date (val) { return this._date = val; },

  get description () { return this._description; },
  set description (val) { return this._description = val; },

  get source () { return this._source; },
  set source (val) { return this._source = val; }
};
